@@ -11,7 +11,7 @@ import (
 	"go/ast"
 	"go/token"
 	"go/types"
-	"log"
+	"path/filepath"
 	"sort"
 	"strings"
 	"time"
@@ -19,9 +19,14 @@ import (
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/lsp/protocol"
 	"golang.org/x/tools/internal/lsp/source"
+	"golang.org/x/tools/internal/lsp/template"
 	errors "golang.org/x/xerrors"
 )
 
+// The LSP says that errors for the semantic token requests should only be returned
+// for exceptions (a word not otherwise defined). This code treats a too-large file
+// as an exception. On parse errors, the code does what it can.
+
 // reject full semantic token requests for large files
 const maxFullFileSize int = 100000
 
@@ -48,7 +53,7 @@ func (s *Server) computeSemanticTokens(ctx context.Context, td protocol.TextDocu
 	ans := protocol.SemanticTokens{
 		Data: []uint32{},
 	}
-	snapshot, _, ok, release, err := s.beginFileRequest(ctx, td.URI, source.Go)
+	snapshot, fh, ok, release, err := s.beginFileRequest(ctx, td.URI, source.UnknownKind)
 	defer release()
 	if !ok {
 		return nil, err
@@ -59,21 +64,39 @@ func (s *Server) computeSemanticTokens(ctx context.Context, td protocol.TextDocu
 		// the client won't remember the wrong answer
 		return nil, errors.Errorf("semantictokens are disabled")
 	}
-	pkg, err := snapshot.PackageForFile(ctx, td.URI.SpanURI(), source.TypecheckFull, source.WidestPackage)
+	if fh.Kind() == source.Tmpl {
+		// this is a little cumbersome to avoid both exporting 'encoded' and its methods
+		// and to avoid import cycles
+		e := &encoded{
+			ctx:      ctx,
+			rng:      rng,
+			tokTypes: s.session.Options().SemanticTypes,
+			tokMods:  s.session.Options().SemanticMods,
+		}
+		add := func(line, start uint32, len uint32) {
+			e.add(line, start, len, tokMacro, nil)
+		}
+		data := func() []uint32 {
+			return e.Data()
+		}
+		return template.SemanticTokens(ctx, snapshot, fh.URI(), add, data)
+	}
+	if fh.Kind() != source.Go {
+		return nil, nil
+	}
+	pkg, err := snapshot.PackageForFile(ctx, fh.URI(), source.TypecheckFull, source.WidestPackage)
 	if err != nil {
 		return nil, err
 	}
 	info := pkg.GetTypesInfo()
-	pgf, err := pkg.File(td.URI.SpanURI())
+	pgf, err := pkg.File(fh.URI())
 	if err != nil {
 		return nil, err
 	}
-	if pgf.ParseErr != nil {
-		return nil, pgf.ParseErr
-	}
+	// ignore pgf.ParseErr. Do what we can.
 	if rng == nil && len(pgf.Src) > maxFullFileSize {
 		err := fmt.Errorf("semantic tokens: file %s too large for full (%d>%d)",
-			td.URI.SpanURI().Filename(), len(pgf.Src), maxFullFileSize)
+			fh.URI().Filename(), len(pgf.Src), maxFullFileSize)
 		return nil, err
 	}
 	e := &encoded{
@@ -86,22 +109,20 @@ func (s *Server) computeSemanticTokens(ctx context.Context, td protocol.TextDocu
 		tokMods:  s.session.Options().SemanticMods,
 	}
 	if err := e.init(); err != nil {
+		// e.init should never return an error, unless there's some
+		// seemingly impossible race condition
 		return nil, err
 	}
 	e.semantics()
-	ans.Data, err = e.Data()
-	if err != nil {
-		// this is an internal error, likely caused by a typo
-		// for a token or modifier
-		return nil, err
-	}
-	// for small cache, some day. for now, the client ignores this
+	ans.Data = e.Data()
+	// For delta requests, but we've never seen any.
 	ans.ResultID = fmt.Sprintf("%v", time.Now())
 	return &ans, nil
 }
 
 func (e *encoded) semantics() {
 	f := e.pgf.File
+	// may not be in range, but harmless
 	e.token(f.Package, len("package"), tokKeyword, nil)
 	e.token(f.Name.NamePos, len(f.Name.Name), tokNamespace, nil)
 	inspect := func(n ast.Node) bool {
@@ -115,6 +136,15 @@ func (e *encoded) semantics() {
 		}
 		ast.Inspect(d, inspect)
 	}
+	for _, cg := range f.Comments {
+		for _, c := range cg.List {
+			if !strings.Contains(c.Text, "\n") {
+				e.token(c.Pos(), len(c.Text), tokComment, nil)
+				continue
+			}
+			e.multiline(c.Pos(), c.End(), c.Text, tokComment)
+		}
+	}
 }
 
 type tokenType string
@@ -132,11 +162,16 @@ const (
 	tokString    tokenType = "string"
 	tokNumber    tokenType = "number"
 	tokOperator  tokenType = "operator"
+
+	tokMacro tokenType = "macro" // for templates
 )
 
 func (e *encoded) token(start token.Pos, leng int, typ tokenType, mods []string) {
-	if start == 0 {
-		e.unexpected("token at token.NoPos")
+
+	if !start.IsValid() {
+		// This is not worth reporting
+		//e.unexpected("token at token.NoPos")
+		return
 	}
 	if start >= e.end || start+token.Pos(leng) <= e.start {
 		return
@@ -150,14 +185,12 @@ func (e *encoded) token(start token.Pos, leng int, typ tokenType, mods []string)
 		// "column mapper is for file...instead of..."
 		// "line is beyond end of file..."
 		// see line 116 of internal/span/token.go which uses Position not PositionFor
-		event.Error(e.ctx, "failed to convert to range", err)
+		// (it is too verbose to print the error on every token. some other RPC will fail)
+		// event.Error(e.ctx, "failed to convert to range", err)
 		return
 	}
 	if lspRange.End.Line != lspRange.Start.Line {
-		// abrupt end of file, without \n. TODO(pjw): fix?
-		pos := e.fset.PositionFor(start, false)
-		msg := fmt.Sprintf("token at %s:%d.%d overflows", pos.Filename, pos.Line, pos.Column)
-		event.Log(e.ctx, msg)
+		// this happens if users are typing at the end of the file, but report nothing
 		return
 	}
 	// token is all on one line
@@ -204,12 +237,26 @@ func (e *encoded) strStack() string {
 	if len(e.stack) > 0 {
 		loc := e.stack[len(e.stack)-1].Pos()
 		add := e.pgf.Tok.PositionFor(loc, false)
-		msg = append(msg, fmt.Sprintf("(line:%d,col:%d)", add.Line, add.Column))
+		nm := filepath.Base(add.Filename)
+		msg = append(msg, fmt.Sprintf("(%s:%d,col:%d)", nm, add.Line, add.Column))
 	}
 	msg = append(msg, "]")
 	return strings.Join(msg, " ")
 }
 
+// find the line in the source
+func (e *encoded) srcLine(x ast.Node) string {
+	file := e.pgf.Tok
+	line := file.Line(x.Pos())
+	start := file.Offset(file.LineStart(line))
+	end := start
+	for ; end < len(e.pgf.Src) && e.pgf.Src[end] != '\n'; end++ {
+
+	}
+	ans := e.pgf.Src[start:end]
+	return string(ans)
+}
+
 func (e *encoded) inspector(n ast.Node) bool {
 	pop := func() {
 		e.stack = e.stack[:len(e.stack)-1]
@@ -224,9 +271,9 @@ func (e *encoded) inspector(n ast.Node) bool {
 	case *ast.AssignStmt:
 		e.token(x.TokPos, len(x.Tok.String()), tokOperator, nil)
 	case *ast.BasicLit:
-		// if it extends across a line, skip it
-		// better would be to mark each line as string TODO(pjw)
 		if strings.Contains(x.Value, "\n") {
+			// has to be a string
+			e.multiline(x.Pos(), x.End(), x.Value, tokString)
 			break
 		}
 		ln := len(x.Value)
@@ -349,11 +396,13 @@ func (e *encoded) inspector(n ast.Node) bool {
 	case *ast.UnaryExpr:
 		e.token(x.OpPos, len(x.Op.String()), tokOperator, nil)
 	case *ast.ValueSpec:
-	// things we won't see
-	case *ast.BadDecl, *ast.BadExpr, *ast.BadStmt,
-		*ast.File, *ast.Package:
-		log.Printf("implement %T %s", x, e.pgf.Tok.PositionFor(x.Pos(), false))
-	// things we knowingly ignore
+	// things only seen with parsing or type errors, so ignore them
+	case *ast.BadDecl, *ast.BadExpr, *ast.BadStmt:
+		return true
+	// not going to see these
+	case *ast.File, *ast.Package:
+		e.unexpected(fmt.Sprintf("implement %T %s", x, e.pgf.Tok.PositionFor(x.Pos(), false)))
+	// other things we knowingly ignore
 	case *ast.Comment, *ast.CommentGroup:
 		pop()
 		return false
@@ -362,6 +411,7 @@ func (e *encoded) inspector(n ast.Node) bool {
 	}
 	return true
 }
+
 func (e *encoded) ident(x *ast.Ident) {
 	def := e.ti.Defs[x]
 	if def != nil {
@@ -374,7 +424,8 @@ func (e *encoded) ident(x *ast.Ident) {
 	use := e.ti.Uses[x]
 	switch y := use.(type) {
 	case nil:
-		e.token(x.NamePos, len(x.Name), tokVariable, []string{"definition"})
+		e.unkIdent(x)
+		return
 	case *types.Builtin:
 		e.token(x.NamePos, len(x.Name), tokFunction, []string{"defaultLibrary"})
 	case *types.Const:
@@ -414,7 +465,7 @@ func (e *encoded) ident(x *ast.Ident) {
 	case *types.Var:
 		e.token(x.Pos(), len(x.Name), tokVariable, nil)
 	default:
-		// replace with panic after extensive testing
+		// can't happen
 		if use == nil {
 			msg := fmt.Sprintf("%#v/%#v %#v %#v", x, x.Obj, e.ti.Defs[x], e.ti.Uses[x])
 			e.unexpected(msg)
@@ -427,6 +478,134 @@ func (e *encoded) ident(x *ast.Ident) {
 	}
 }
 
+// both e.ti.Defs and e.ti.Uses are nil. use the parse stack
+// a lot of these only happen when the package doesn't compile
+func (e *encoded) unkIdent(x *ast.Ident) {
+	tok := func(tok tokenType, mod []string) {
+		e.token(x.Pos(), len(x.Name), tok, mod)
+	}
+	def := []string{"definition"}
+	n := len(e.stack) - 2 // parent of Ident
+	if n < 0 {
+		e.unexpected("no stack?")
+		return
+	}
+	switch nd := e.stack[n].(type) {
+	case *ast.BinaryExpr, *ast.UnaryExpr, *ast.ParenExpr, *ast.StarExpr,
+		*ast.IncDecStmt, *ast.SliceExpr, *ast.ExprStmt, *ast.IndexExpr,
+		*ast.ReturnStmt,
+		*ast.IfStmt,       /* condition */
+		*ast.KeyValueExpr: // either key or value
+		tok(tokVariable, nil)
+	case *ast.Ellipsis:
+		tok(tokType, nil)
+	case *ast.CaseClause:
+		if n-2 >= 0 {
+			if _, ok := e.stack[n-2].(*ast.TypeSwitchStmt); ok {
+				tok(tokType, nil)
+				return
+			}
+		}
+		tok(tokVariable, nil)
+	case *ast.ArrayType:
+		if x == nd.Len {
+			tok(tokVariable, nil)
+		} else {
+			tok(tokType, nil)
+		}
+	case *ast.MapType:
+		tok(tokType, nil)
+	case *ast.CallExpr:
+		if x == nd.Fun {
+			tok(tokFunction, nil)
+			return
+		}
+		tok(tokVariable, nil)
+	case *ast.TypeAssertExpr:
+		if x == nd.X {
+			tok(tokVariable, nil)
+		} else if x == nd.Type {
+			tok(tokType, nil)
+		}
+	case *ast.ValueSpec:
+		for _, p := range nd.Names {
+			if p == x {
+				tok(tokVariable, def)
+				return
+			}
+		}
+		for _, p := range nd.Values {
+			if p == x {
+				tok(tokVariable, nil)
+				return
+			}
+		}
+		tok(tokType, nil)
+	case *ast.SelectorExpr: // e.ti.Selections[nd] is nil, so no help
+		if n-1 >= 0 {
+			if ce, ok := e.stack[n-1].(*ast.CallExpr); ok {
+				// ... CallExpr SelectorExpr Ident (_.x())
+				if ce.Fun == nd && nd.Sel == x {
+					tok(tokFunction, nil)
+					return
+				}
+			}
+		}
+		tok(tokVariable, nil)
+	case *ast.AssignStmt:
+		for _, p := range nd.Lhs {
+			// x := ..., or x = ...
+			if p == x {
+				if nd.Tok != token.DEFINE {
+					def = nil
+				}
+				tok(tokVariable, def)
+				return
+			}
+		}
+		// RHS, = x
+		tok(tokVariable, nil)
+	case *ast.TypeSpec: // it's a type if it is either the Name or the Type
+		if x == nd.Type {
+			def = nil
+		}
+		tok(tokType, def)
+	case *ast.Field:
+		// ident could be type in a field, or a method in an interface type, or a variable
+		if x == nd.Type {
+			tok(tokType, nil)
+			return
+		}
+		if n-2 >= 0 {
+			_, okit := e.stack[n-2].(*ast.InterfaceType)
+			_, okfl := e.stack[n-1].(*ast.FieldList)
+			if okit && okfl {
+				tok(tokMember, def)
+				return
+			}
+		}
+		tok(tokVariable, nil)
+	case *ast.LabeledStmt, *ast.BranchStmt:
+		// nothing to report
+	case *ast.CompositeLit:
+		if nd.Type == x {
+			tok(tokType, nil)
+			return
+		}
+		tok(tokVariable, nil)
+	case *ast.RangeStmt:
+		if nd.Tok != token.DEFINE {
+			def = nil
+		}
+		tok(tokVariable, def)
+	case *ast.FuncDecl:
+		tok(tokFunction, def)
+	default:
+		msg := fmt.Sprintf("%T undexpected: %s %s%q", nd, x.Name, e.strStack(), e.srcLine(x))
+		e.unexpected(msg)
+	}
+}
+
 func isDeprecated(n *ast.CommentGroup) bool {
 	if n == nil {
 		return false
@@ -481,20 +660,49 @@ func (e *encoded) definitionFor(x *ast.Ident) (tokenType, []string) {
 			// (type A func(b uint64)) (err error)
 			// b and err should not be tokType, but tokVaraible
 			// and in GenDecl/TpeSpec/StructType/FieldList/Field/Ident
-			// (type A struct{b uint64})
+			// (type A struct{b uint64}
+			// but on type B struct{C}), C is a type, but is not being defined.
 			fldm := e.stack[len(e.stack)-2]
-			if _, ok := fldm.(*ast.Field); ok {
+			if fld, ok := fldm.(*ast.Field); ok {
+				// if len(fld.names) == 0 this is a tokType, being used
+				if len(fld.Names) == 0 {
+					return tokType, nil
+				}
 				return tokVariable, mods
 			}
 			return tokType, mods
 		}
 	}
-	// panic after extensive testing
+	// can't happen
 	msg := fmt.Sprintf("failed to find the decl for %s", e.pgf.Tok.PositionFor(x.Pos(), false))
 	e.unexpected(msg)
 	return "", []string{""}
 }
 
+func (e *encoded) multiline(start, end token.Pos, val string, tok tokenType) {
+	f := e.fset.File(start)
+	// the hard part is finding the lengths of lines. include the \n
+	leng := func(line int) int {
+		n := f.LineStart(line)
+		if line >= f.LineCount() {
+			return f.Size() - int(n)
+		}
+		return int(f.LineStart(line+1) - n)
+	}
+	spos := e.fset.PositionFor(start, false)
+	epos := e.fset.PositionFor(end, false)
+	sline := spos.Line
+	eline := epos.Line
+	// first line is from spos.Column to end
+	e.token(start, leng(sline)-spos.Column, tok, nil) // leng(sline)-1 - (spos.Column-1)
+	for i := sline + 1; i < eline; i++ {
+		// intermediate lines are from 1 to end
+		e.token(f.LineStart(i), leng(i)-1, tok, nil) // avoid the newline
+	}
+	// last line is from 1 to epos.Column
+	e.token(f.LineStart(eline), epos.Column-1, tok, nil) // columns are 1-based
+}
+
 // findKeyword finds a keyword rather than guessing its location
 func (e *encoded) findKeyword(keyword string, start, end token.Pos) token.Pos {
 	offset := int(start) - e.pgf.Tok.Base()
@@ -517,14 +725,14 @@ func (e *encoded) init() error {
 	}
 	span, err := e.pgf.Mapper.RangeSpan(*e.rng)
 	if err != nil {
-		return errors.Errorf("range span (%v) error for %s", err, e.pgf.File.Name)
+		return errors.Errorf("range span (%w) error for %s", err, e.pgf.File.Name)
 	}
 	e.end = e.start + token.Pos(span.End().Offset())
 	e.start += token.Pos(span.Start().Offset())
 	return nil
 }
 
-func (e *encoded) Data() ([]uint32, error) {
+func (e *encoded) Data() []uint32 {
 	// binary operators, at least, will be out of order
 	sort.Slice(e.items, func(i, j int) bool {
 		if e.items[i].line != e.items[j].line {
@@ -548,14 +756,19 @@ func (e *encoded) Data() ([]uint32, error) {
 			x[j+1] = e.items[i].start - e.items[i-1].start
 		}
 		x[j+2] = e.items[i].len
-		x[j+3] = uint32(typeMap[e.items[i].typeStr])
+		typ, ok := typeMap[e.items[i].typeStr]
+		if !ok {
+			continue // client doesn't want typeStr
+		}
+		x[j+3] = uint32(typ)
 		mask := 0
 		for _, s := range e.items[i].mods {
+			// modMpa[s] is 0 if the client doesn't want this modifier
 			mask |= modMap[s]
 		}
 		x[j+4] = uint32(mask)
 	}
-	return x, nil
+	return x
 }
 
 func (e *encoded) importSpec(d *ast.ImportSpec) {
@@ -573,20 +786,22 @@ func (e *encoded) importSpec(d *ast.ImportSpec) {
 		}
 		// and fall through for _
 	}
-	nm := d.Path.Value[1 : len(d.Path.Value)-1] // trailing "
-	v := strings.LastIndex(nm, "/")
-	if v != -1 {
-		nm = nm[v+1:]
+	val := d.Path.Value
+	if len(val) < 2 || val[0] != '"' || val[len(val)-1] != '"' {
+		// avoid panics on imports without a properly quoted string
+		return
 	}
+	nm := val[1 : len(val)-1] // remove surrounding "s
+	nm = filepath.Base(nm)
+	// in import "lib/math", 'math' is the package name
 	start := d.Path.End() - token.Pos(1+len(nm))
 	e.token(start, len(nm), tokNamespace, nil)
+	// There may be more cases, as import strings are implementation defined.
 }
 
-// panic on unexpected state
+// log unexpected state
 func (e *encoded) unexpected(msg string) {
-	log.Print(msg)
-	log.Print(e.strStack())
-	panic(msg)
+	event.Error(e.ctx, e.strStack(), errors.New(msg))
 }
 
 // SemType returns a string equivalent of the type, for gopls semtok
@@ -638,8 +853,10 @@ var (
 		"namespace", "type", "class", "enum", "interface",
 		"struct", "typeParameter", "parameter", "variable", "property", "enumMember",
 		"event", "function", "member", "macro", "keyword", "modifier", "comment",
-		"string", "number", "regexp", "operator"}
+		"string", "number", "regexp", "operator",
+	}
 	semanticModifiers = [...]string{
 		"declaration", "definition", "readonly", "static",
-		"deprecated", "abstract", "async", "modification", "documentation", "defaultLibrary"}
+		"deprecated", "abstract", "async", "modification", "documentation", "defaultLibrary",
+	}
 )
