@@ -25,13 +25,13 @@ import (
 // sent in response to a client.
 const maxSymbols = 100
 
-// WorkspaceSymbols matches symbols across views using the given query,
-// according to the SymbolMatcher matcher.
+// WorkspaceSymbols matches symbols across all views using the given query,
+// according to the match semantics parameterized by matcherType and style.
 //
 // The workspace symbol method is defined in the spec as follows:
 //
-//  > The workspace symbol request is sent from the client to the server to
-//  > list project-wide symbols matching the query string.
+//   The workspace symbol request is sent from the client to the server to
+//   list project-wide symbols matching the query string.
 //
 // It is unclear what "project-wide" means here, but given the parameters of
 // workspace/symbol do not include any workspace identifier, then it has to be
@@ -57,79 +57,77 @@ func WorkspaceSymbols(ctx context.Context, matcherType SymbolMatcher, style Symb
 // See the comment for symbolCollector for more information.
 type matcherFunc func(name string) float64
 
-// A symbolizer returns the best symbol match for name with pkg, according to
-// some heuristic.
+// A symbolizer returns the best symbol match for a name with pkg, according to
+// some heuristic. The symbol name is passed as the slice nameParts of logical
+// name pieces. For example, for myType.field the caller can pass either
+// []string{"myType.field"} or []string{"myType.", "field"}.
 //
 // See the comment for symbolCollector for more information.
-type symbolizer func(name string, pkg Package, m matcherFunc) (string, float64)
+type symbolizer func(nameParts []string, pkg Package, m matcherFunc) (string, float64)
 
-func fullyQualifiedSymbolMatch(name string, pkg Package, matcher matcherFunc) (string, float64) {
-	fullyQualified := pkg.PkgPath() + "." + name
-	if matcher(fullyQualified) > 0 {
-		return fullyQualified, 1
+func fullyQualifiedSymbolMatch(nameParts []string, pkg Package, matcher matcherFunc) (string, float64) {
+	_, score := dynamicSymbolMatch(nameParts, pkg, matcher)
+	path := append([]string{pkg.PkgPath() + "."}, nameParts...)
+	if score > 0 {
+		return strings.Join(path, ""), score
 	}
 	return "", 0
 }
 
-func dynamicSymbolMatch(name string, pkg Package, matcher matcherFunc) (string, float64) {
-	// Prefer any package-qualified match.
-	pkgQualified := pkg.Name() + "." + name
-	if match, score := bestMatch(pkgQualified, matcher); match != "" {
-		return match, score
+func dynamicSymbolMatch(nameParts []string, pkg Package, matcher matcherFunc) (string, float64) {
+	var best string
+	fullName := strings.Join(nameParts, "")
+	var score float64
+	var name string
+
+	// Compute the match score by finding the highest scoring suffix. In these
+	// cases the matched symbol is still the full name: it is confusing to match
+	// an unqualified nested field or method.
+	if match := bestMatch("", nameParts, matcher); match > score {
+		best = fullName
+		score = match
 	}
-	fullyQualified := pkg.PkgPath() + "." + name
-	if match, score := bestMatch(fullyQualified, matcher); match != "" {
-		return match, score
+
+	// Next: try to match a package-qualified name.
+	name = pkg.Name() + "." + fullName
+	if match := matcher(name); match > score {
+		best = name
+		score = match
 	}
-	return "", 0
+
+	// Finally: consider a fully qualified name.
+	prefix := pkg.PkgPath() + "."
+	fullyQualified := prefix + fullName
+	// As with field/method selectors, consider suffixes from right to left, but
+	// always return a fully-qualified symbol.
+	pathParts := strings.SplitAfter(prefix, "/")
+	if match := bestMatch(fullName, pathParts, matcher); match > score {
+		best = fullyQualified
+		score = match
+	}
+	return best, score
 }
 
-func packageSymbolMatch(name string, pkg Package, matcher matcherFunc) (string, float64) {
-	qualified := pkg.Name() + "." + name
+func bestMatch(name string, prefixParts []string, matcher matcherFunc) float64 {
+	var score float64
+	for i := len(prefixParts) - 1; i >= 0; i-- {
+		name = prefixParts[i] + name
+		if match := matcher(name); match > score {
+			score = match
+		}
+	}
+	return score
+}
+
+func packageSymbolMatch(components []string, pkg Package, matcher matcherFunc) (string, float64) {
+	path := append([]string{pkg.Name() + "."}, components...)
+	qualified := strings.Join(path, "")
 	if matcher(qualified) > 0 {
 		return qualified, 1
 	}
 	return "", 0
 }
 
-// bestMatch returns the highest scoring symbol suffix of fullPath, starting
-// from the right and splitting on selectors and path components.
-//
-// e.g. given a symbol path of the form 'host.com/dir/pkg.type.field', we
-// check the match quality of the following:
-//  - field
-//  - type.field
-//  - pkg.type.field
-//  - dir/pkg.type.field
-//  - host.com/dir/pkg.type.field
-//
-// and return the best match, along with its score.
-//
-// This is used to implement the 'dynamic' symbol style.
-func bestMatch(fullPath string, matcher matcherFunc) (string, float64) {
-	pathParts := strings.Split(fullPath, "/")
-	dottedParts := strings.Split(pathParts[len(pathParts)-1], ".")
-
-	var best string
-	var score float64
-
-	for i := 0; i < len(dottedParts); i++ {
-		path := strings.Join(dottedParts[len(dottedParts)-1-i:], ".")
-		if match := matcher(path); match > score {
-			best = path
-			score = match
-		}
-	}
-	for i := 0; i < len(pathParts); i++ {
-		path := strings.Join(pathParts[len(pathParts)-1-i:], "/")
-		if match := matcher(path); match > score {
-			best = path
-			score = match
-		}
-	}
-	return best, score
-}
-
 // symbolCollector holds context as we walk Packages, gathering symbols that
 // match a given query.
 //
@@ -140,9 +138,6 @@ func bestMatch(fullPath string, matcher matcherFunc) (string, float64) {
 //  * A symbolizer determines how we extract the symbol for an object. This
 //    enables the 'symbolStyle' configuration option.
 type symbolCollector struct {
-	// query is the user-supplied query passed to the Symbol method.
-	query string
-
 	// These types parameterize the symbol-matching pass.
 	matcher    matcherFunc
 	symbolizer symbolizer
@@ -274,8 +269,7 @@ func (c comboMatcher) match(s string) float64 {
 
 // walk walks views, gathers symbols, and returns the results.
 func (sc *symbolCollector) walk(ctx context.Context, views []View) (_ []protocol.SymbolInformation, err error) {
-	toWalk, release, err := sc.collectPackages(ctx, views)
-	defer release()
+	toWalk, err := sc.collectPackages(ctx, views)
 	if err != nil {
 		return nil, err
 	}
@@ -307,79 +301,29 @@ func (sc *symbolCollector) results() []protocol.SymbolInformation {
 	return res
 }
 
-// collectPackages gathers the packages we are going to inspect for symbols.
-// This pre-step is required in order to filter out any "duplicate"
-// *types.Package. The duplicates arise for packages that have test variants.
-// For example, if package mod.com/p has test files, then we will visit two
-// packages that have the PkgPath() mod.com/p: the first is the actual package
-// mod.com/p, the second is a special version that includes the non-XTest
-// _test.go files. If we were to walk both of of these packages, then we would
-// get duplicate matching symbols and we would waste effort. Therefore where
-// test variants exist we walk those (because they include any symbols defined
-// in non-XTest _test.go files).
-//
-// One further complication is that even after this filtering, packages between
-// views might not be "identical" because they can be built using different
-// build constraints (via the "env" config option).
-//
-// Therefore on a per view basis we first build up a map of package path ->
-// *types.Package preferring the test variants if they exist. Then we merge the
-// results between views, de-duping by *types.Package.
-func (sc *symbolCollector) collectPackages(ctx context.Context, views []View) ([]*pkgView, func(), error) {
-	gathered := make(map[string]map[*types.Package]*pkgView)
-	var releaseFuncs []func()
-	release := func() {
-		for _, releaseFunc := range releaseFuncs {
-			releaseFunc()
-		}
-	}
+// collectPackages gathers all known packages and sorts for stability.
+func (sc *symbolCollector) collectPackages(ctx context.Context, views []View) ([]*pkgView, error) {
 	var toWalk []*pkgView
 	for _, v := range views {
-		seen := make(map[string]*pkgView)
 		snapshot, release := v.Snapshot(ctx)
-		releaseFuncs = append(releaseFuncs, release)
+		defer release()
 		knownPkgs, err := snapshot.KnownPackages(ctx)
 		if err != nil {
-			return nil, release, err
+			return nil, err
 		}
 		workspacePackages, err := snapshot.WorkspacePackages(ctx)
 		if err != nil {
-			return nil, release, err
+			return nil, err
 		}
 		isWorkspacePkg := make(map[Package]bool)
 		for _, wp := range workspacePackages {
 			isWorkspacePkg[wp] = true
 		}
-		var forTests []*pkgView
 		for _, pkg := range knownPkgs {
-			toAdd := &pkgView{
+			toWalk = append(toWalk, &pkgView{
 				pkg:         pkg,
-				snapshot:    snapshot,
 				isWorkspace: isWorkspacePkg[pkg],
-			}
-			// Defer test packages, so that they overwrite seen for this package
-			// path.
-			if pkg.ForTest() != "" {
-				forTests = append(forTests, toAdd)
-			} else {
-				seen[pkg.PkgPath()] = toAdd
-			}
-		}
-		for _, pkg := range forTests {
-			seen[pkg.pkg.PkgPath()] = pkg
-		}
-		for _, pkg := range seen {
-			pm, ok := gathered[pkg.pkg.PkgPath()]
-			if !ok {
-				pm = make(map[*types.Package]*pkgView)
-				gathered[pkg.pkg.PkgPath()] = pm
-			}
-			pm[pkg.pkg.GetTypes()] = pkg
-		}
-	}
-	for _, pm := range gathered {
-		for _, pkg := range pm {
-			toWalk = append(toWalk, pkg)
+			})
 		}
 	}
 	// Now sort for stability of results. We order by
@@ -396,7 +340,7 @@ func (sc *symbolCollector) collectPackages(ctx context.Context, views []View) ([
 			return false
 		}
 	})
-	return toWalk, release, nil
+	return toWalk, nil
 }
 
 func (sc *symbolCollector) walkFilesDecls(decls []ast.Decl) {
@@ -405,14 +349,9 @@ func (sc *symbolCollector) walkFilesDecls(decls []ast.Decl) {
 		case *ast.FuncDecl:
 			kind := protocol.Function
 			var recv *ast.Ident
-			if decl.Recv != nil {
+			if decl.Recv.NumFields() > 0 {
 				kind = protocol.Method
-				switch typ := decl.Recv.List[0].Type.(type) {
-				case *ast.StarExpr:
-					recv = typ.X.(*ast.Ident)
-				case *ast.Ident:
-					recv = typ
-				}
+				recv = unpackRecv(decl.Recv.List[0].Type)
 			}
 			if recv != nil {
 				sc.match(decl.Name.Name, kind, decl.Name, recv)
@@ -439,6 +378,25 @@ func (sc *symbolCollector) walkFilesDecls(decls []ast.Decl) {
 	}
 }
 
+func unpackRecv(rtyp ast.Expr) *ast.Ident {
+	// Extract the receiver identifier. Lifted from go/types/resolver.go
+L:
+	for {
+		switch t := rtyp.(type) {
+		case *ast.ParenExpr:
+			rtyp = t.X
+		case *ast.StarExpr:
+			rtyp = t.X
+		default:
+			break L
+		}
+	}
+	if name, _ := rtyp.(*ast.Ident); name != nil {
+		return name
+	}
+	return nil
+}
+
 // walkType processes symbols related to a type expression. path is path of
 // nested type identifiers to the type expression.
 func (sc *symbolCollector) walkType(typ ast.Expr, path ...*ast.Ident) {
@@ -460,7 +418,13 @@ func (sc *symbolCollector) walkType(typ ast.Expr, path ...*ast.Ident) {
 // or named. path is the path of nested identifiers containing the field.
 func (sc *symbolCollector) walkField(field *ast.Field, unnamedKind, namedKind protocol.SymbolKind, path ...*ast.Ident) {
 	if len(field.Names) == 0 {
-		sc.match(types.ExprString(field.Type), unnamedKind, field, path...)
+		switch typ := field.Type.(type) {
+		case *ast.SelectorExpr:
+			// embedded qualified type
+			sc.match(typ.Sel.Name, unnamedKind, field, path...)
+		default:
+			sc.match(types.ExprString(field.Type), unnamedKind, field, path...)
+		}
 	}
 	for _, name := range field.Names {
 		sc.match(name.Name, namedKind, name, path...)
@@ -506,18 +470,14 @@ func (sc *symbolCollector) match(name string, kind protocol.SymbolKind, node ast
 	}
 
 	isExported := isExported(name)
-	if len(path) > 0 {
-		var nameBuilder strings.Builder
-		for _, ident := range path {
-			nameBuilder.WriteString(ident.Name)
-			nameBuilder.WriteString(".")
-			if !ident.IsExported() {
-				isExported = false
-			}
+	var names []string
+	for _, ident := range path {
+		names = append(names, ident.Name+".")
+		if !ident.IsExported() {
+			isExported = false
 		}
-		nameBuilder.WriteString(name)
-		name = nameBuilder.String()
 	}
+	names = append(names, name)
 
 	// Factors to apply to the match score for the purpose of downranking
 	// results.
@@ -541,7 +501,7 @@ func (sc *symbolCollector) match(name string, kind protocol.SymbolKind, node ast
 		// can be noisy.
 		fieldFactor = 0.5
 	)
-	symbol, score := sc.symbolizer(name, sc.current.pkg, sc.matcher)
+	symbol, score := sc.symbolizer(names, sc.current.pkg, sc.matcher)
 
 	// Downrank symbols outside of the workspace.
 	if !sc.current.isWorkspace {
@@ -562,8 +522,7 @@ func (sc *symbolCollector) match(name string, kind protocol.SymbolKind, node ast
 		return
 	}
 
-	mrng := NewMappedRange(sc.current.snapshot.FileSet(), sc.curFile.Mapper, node.Pos(), node.End())
-	rng, err := mrng.Range()
+	rng, err := fileRange(sc.curFile, node.Pos(), node.End())
 	if err != nil {
 		return
 	}
@@ -574,7 +533,7 @@ func (sc *symbolCollector) match(name string, kind protocol.SymbolKind, node ast
 		container: sc.current.pkg.PkgPath(),
 		kind:      kind,
 		location: protocol.Location{
-			URI:   protocol.URIFromSpanURI(mrng.URI()),
+			URI:   protocol.URIFromSpanURI(sc.curFile.URI),
 			Range: rng,
 		},
 	}
@@ -587,6 +546,14 @@ func (sc *symbolCollector) match(name string, kind protocol.SymbolKind, node ast
 	sc.res[insertAt] = si
 }
 
+func fileRange(pgf *ParsedGoFile, start, end token.Pos) (protocol.Range, error) {
+	s, err := span.FileSpan(pgf.Tok, pgf.Mapper.Converter, start, end)
+	if err != nil {
+		return protocol.Range{}, nil
+	}
+	return pgf.Mapper.Range(s)
+}
+
 // isExported reports if a token is exported. Copied from
 // token.IsExported (go1.13+).
 //
@@ -600,7 +567,6 @@ func isExported(name string) bool {
 // pkgView holds information related to a package that we are going to walk.
 type pkgView struct {
 	pkg         Package
-	snapshot    Snapshot
 	isWorkspace bool
 }
 
